@@ -34,8 +34,6 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gAddressBookBundle;
-
 function GetAbViewListener()
 {
   // the ab panel doesn't care if the total changes, or if the selection changes
@@ -134,10 +132,20 @@ function AddressBookMenuListChange()
   }
 }
 
+function AbPanelOnComposerClose()
+{
+  CloseAbView();
+  onClearSearch();
+}
+
+function AbPanelOnComposerReOpen()
+{
+  ChangeDirectoryByURI(GetSelectedDirectory());
+}
+
 function AbPanelLoad() 
 {
   InitCommonJS();
-  gAddressBookBundle = document.getElementById("bundle_addressBook");
 
   LoadPreviouslySelectedAB();
 
@@ -150,7 +158,9 @@ function AbPanelLoad()
     gAddressBookPanelAbListener,
     Components.interfaces.nsIAddrBookSession.directoryRemoved |
     Components.interfaces.nsIAddrBookSession.changed);
-
+  
+  parent.document.getElementById("msgcomposeWindow").addEventListener('compose-window-close', AbPanelOnComposerClose, true);
+  parent.document.getElementById("msgcomposeWindow").addEventListener('compose-window-reopen', AbPanelOnComposerReOpen, true);
   gSearchInput = document.getElementById("searchInput");
 }
 
@@ -158,6 +168,8 @@ function AbPanelUnload()
 {
   var addrbookSession = Components.classes["@mozilla.org/addressbook/services/session;1"].getService().QueryInterface(Components.interfaces.nsIAddrBookSession);
   addrbookSession.removeAddressBookListener(gAddressBookPanelAbListener);
+  parent.document.getElementById("msgcomposeWindow").removeEventListener('compose-window-close', AbPanelOnComposerClose, true);
+  parent.document.getElementById("msgcomposeWindow").removeEventListener('compose-window-reopen', AbPanelOnComposerReOpen, true);
 
   CloseAbView();
 }
@@ -206,5 +218,5 @@ function onEnterInSearchBar()
   if (gSearchInput.value != "")
     searchURI += gQueryURIFormat.replace(/@V/g, encodeURIComponent(gSearchInput.value));
 
-  SetAbView(searchURI, true, sortColumn, sortDirection);
+  SetAbView(searchURI, gSearchInput.value != "", sortColumn, sortDirection);
 }
