@@ -50,24 +50,25 @@
 ------------------------------------------------------------------------------*/
 
 function nsContextMenu( xulMenu ) {
-    this.target         = null;
-    this.menu           = null;
-    this.popupURL       = null;
-    this.onTextInput    = false;
-    this.onImage        = false;
-    this.onLoadedImage  = false;
-    this.onLink         = false;
-    this.onMailtoLink   = false;
-    this.onSaveableLink = false;
-    this.onMetaDataItem = false;
-    this.onMathML       = false;
-    this.link           = false;
-    this.inFrame        = false;
-    this.hasBGImage     = false;
-    this.isTextSelected = false;
-    this.inDirList      = false;
-    this.shouldDisplay  = true;
-    this.autoDownload   = false;
+    this.target            = null;
+    this.menu              = null;
+    this.popupURL          = null;
+    this.onTextInput       = false;
+    this.onImage           = false;
+    this.onLoadedImage     = false;
+    this.onLink            = false;
+    this.onMailtoLink      = false;
+    this.onSaveableLink    = false;
+    this.onMetaDataItem    = false;
+    this.onMathML          = false;
+    this.link              = false;
+    this.inFrame           = false;
+    this.hasBGImage        = false;
+    this.isTextSelected    = false;
+    this.isContentSelected = false;
+    this.inDirList         = false;
+    this.shouldDisplay     = true;
+    this.autoDownload      = false;
 
     // Initialize new menu.
     this.initMenu( xulMenu );
@@ -80,18 +81,20 @@ nsContextMenu.prototype = {
     },
     // Initialize context menu.
     initMenu : function ( popup ) {
+        // Save menu.
+        this.menu = popup;
+
         const xulNS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
         if ( document.popupNode.namespaceURI == xulNS ) {
           this.shouldDisplay = false;
           return;
         }
-        // Save menu.
-        this.menu = popup;
-
         // Get contextual info.
-        this.setTarget( document.popupNode );
-        
+        this.setTarget( document.popupNode, document.popupRangeParent,
+                        document.popupRangeOffset );
+
         this.isTextSelected = this.isTextSelection();
+        this.isContentSelected = this.isContentSelection();
 
         this.initPopupURL();
 
@@ -103,6 +106,7 @@ nsContextMenu.prototype = {
         this.initNavigationItems();
         this.initViewItems();
         this.initMiscItems();
+        this.initSpellingItems();
         this.initSaveItems();
         this.initClipboardItems();
         this.initMetadataItems();
@@ -122,7 +126,7 @@ nsContextMenu.prototype = {
         // Forward determined by canGoForward broadcaster.
         this.setItemAttrFromNode( "context-forward", "disabled", "canGoForward" );
 
-        var showNav = !( this.isTextSelected || this.onLink || this.onImage || this.onTextInput );
+        var showNav = !( this.isContentSelected || this.onLink || this.onImage || this.onTextInput );
         
         this.showItem( "context-back", showNav );
         this.showItem( "context-forward", showNav );
@@ -136,7 +140,7 @@ nsContextMenu.prototype = {
         //this.setItemAttrFromNode( "context-stop", "disabled", "canStop" );
     },
     initSaveItems : function () {
-        var showSave = !( this.inDirList || this.isTextSelected || this.onTextInput || this.onStandaloneImage ||
+        var showSave = !( this.inDirList || this.isContentSelected || this.onTextInput || this.onStandaloneImage ||
                        ( this.onLink && this.onImage ) );
         if (showSave)
           goSetMenuValue( "context-savepage", this.autoDownload ? "valueSave" : "valueSaveAs" );
@@ -156,15 +160,15 @@ nsContextMenu.prototype = {
     },
     initViewItems : function () {
         // View source is always OK, unless in directory listing.
-        this.showItem( "context-viewpartialsource-selection", this.isTextSelected && !this.onTextInput );
-        this.showItem( "context-viewpartialsource-mathml", this.onMathML && !this.isTextSelected );
+        this.showItem( "context-viewpartialsource-selection", this.isContentSelected && !this.onTextInput );
+        this.showItem( "context-viewpartialsource-mathml", this.onMathML && !this.isContentSelected );
 
-        var showView = !( this.inDirList || this.onImage || this.isTextSelected || this.onLink || this.onTextInput );
+        var showView = !( this.inDirList || this.onImage || this.isContentSelected || this.onLink || this.onTextInput );
 
         this.showItem( "context-viewsource", showView );
         this.showItem( "context-viewinfo", showView );
 
-        this.showItem( "context-sep-properties", !( this.inDirList || this.isTextSelected || this.onTextInput ) );
+        this.showItem( "context-sep-properties", !( this.inDirList || this.isContentSelected || this.onTextInput ) );
         // Set As Wallpaper depends on whether an image was clicked on, and only works on Windows.
         var isWin = navigator.appVersion.indexOf("Windows") != -1;
         this.showItem( "context-setWallpaper", isWin && (this.onLoadedImage || this.onStandaloneImage));
@@ -191,7 +195,7 @@ nsContextMenu.prototype = {
     },
     initMiscItems : function () {
         // Use "Bookmark This Link" if on a link.
-        this.showItem( "context-bookmarkpage", !( this.isTextSelected || this.onTextInput || this.onStandaloneImage ) );
+        this.showItem( "context-bookmarkpage", !( this.isContentSelected || this.onTextInput || this.onStandaloneImage ) );
         this.showItem( "context-bookmarklink", this.onLink && !this.onMailtoLink );
         this.showItem( "context-searchselect", this.isTextSelected && !this.onTextInput );
         this.showItem( "frame", this.inFrame );
@@ -217,6 +221,40 @@ nsContextMenu.prototype = {
         this.showItem( "context-bidi-text-direction-toggle", this.onTextInput && gShowBiDi);
         this.showItem( "context-bidi-page-direction-toggle", !this.onTextInput && gShowBiDi);
     },
+    initSpellingItems : function () {
+        var canSpell = InlineSpellCheckerUI.canSpellCheck;
+        var onMisspelling = InlineSpellCheckerUI.overMisspelling;
+        this.showItem("spell-check-enabled", canSpell);
+        this.showItem("spell-separator", canSpell || this.possibleSpellChecking);
+        if (canSpell)
+            document.getElementById("spell-check-enabled").setAttribute("checked",
+                                                                        InlineSpellCheckerUI.enabled);
+        this.showItem("spell-add-to-dictionary", onMisspelling);
+
+        // suggestion list
+        this.showItem("spell-suggestions-separator", onMisspelling);
+        if (onMisspelling) {
+            var menu = document.getElementById("contentAreaContextMenu");
+            var suggestionsSeparator = document.getElementById("spell-add-to-dictionary");
+            var numsug = InlineSpellCheckerUI.addSuggestionsToMenu(menu, suggestionsSeparator, 5);
+            this.showItem("spell-no-suggestions", numsug == 0);
+        } else {
+            this.showItem("spell-no-suggestions", false);
+        }
+
+        // dictionary list
+        this.showItem("spell-dictionaries", InlineSpellCheckerUI.enabled);
+        if (canSpell) {
+            var dictMenu = document.getElementById("spell-dictionaries-menu");
+            var dictSep = document.getElementById("spell-language-separator");
+            InlineSpellCheckerUI.addDictionaryListToMenu(dictMenu, dictSep);
+        }
+
+        // when there is no spellchecker but we might be able to spellcheck
+        // add the add to dictionaries item. This will ensure that people
+        // with no dictionaries will be able to download them
+        this.showItem("spell-add-dictionaries-main", !canSpell && this.possibleSpellChecking);
+    },
     initClipboardItems : function () {
 
         // Copy depends on whether there is selected text.
@@ -230,12 +268,12 @@ nsContextMenu.prototype = {
         this.showItem( "context-redo", this.onTextInput );
         this.showItem( "context-sep-undo", this.onTextInput );
         this.showItem( "context-cut", this.onTextInput );
-        this.showItem( "context-copy", this.isTextSelected || this.onTextInput);
+        this.showItem( "context-copy", this.isContentSelected || this.onTextInput);
         this.showItem( "context-paste", this.onTextInput );
         this.showItem( "context-delete", this.onTextInput );
         this.showItem( "context-sep-paste", this.onTextInput );
-        this.showItem( "context-selectall", true );
-        this.showItem( "context-sep-selectall", this.isTextSelected && !this.onTextInput );
+        this.showItem( "context-selectall", !( this.onLink || this.onImage ) );
+        this.showItem( "context-sep-selectall", this.isContentSelected && !this.onTextInput );
         // In a text area there will be nothing after select all, so we don't want a sep
         // Otherwise, if there's text selected then there are extra menu items
         // (search for selection and view selection source), so we do want a sep
@@ -261,7 +299,7 @@ nsContextMenu.prototype = {
         this.showItem( "context-metadata", this.onMetaDataItem );
     },
     // Set various context menu attributes based on the state of the world.
-    setTarget : function ( node ) {
+    setTarget : function ( node, rangeParent, rangeOffset ) {
         // Initialize contextual info.
         this.onImage    = false;
         this.onLoadedImage = false;
@@ -274,14 +312,44 @@ nsContextMenu.prototype = {
         this.inFrame    = false;
         this.hasBGImage = false;
         this.bgImageURL = "";
+        this.possibleSpellChecking = false;
 
         // Remember the node that was clicked.
         this.target = node;
 
         this.autoDownload = Components.classes["@mozilla.org/preferences-service;1"]
-                                      .getService(Components.interfaces.nsIPrefService)
+                                      .getService(Components.interfaces.nsIPrefBranch)
                                       .getBoolPref("browser.download.autoDownload");
 
+        // Clear any old spellchecking items from the menu, this used to
+        // be in the menu hiding code but wasn't getting called in all
+        // situations. Here, we can ensure it gets cleaned up any time the
+        // menu is shown. Note: must be before uninit because that clears the
+        // internal vars
+        InlineSpellCheckerUI.clearSuggestionsFromMenu();
+        InlineSpellCheckerUI.clearDictionaryListFromMenu();
+
+        InlineSpellCheckerUI.uninit();
+
+        // if the document is editable, show context menu like in text inputs
+        var win = this.target.ownerDocument.defaultView;
+        if (win) {
+          var editingSession = win.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
+                                  .getInterface(Components.interfaces.nsIWebNavigation)
+                                  .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
+                                  .getInterface(Components.interfaces.nsIEditingSession);
+          if (editingSession.windowIsEditable(win)) {
+            this.onTextInput           = true;
+            this.possibleSpellChecking = true;
+            InlineSpellCheckerUI.init(editingSession.getEditorForWindow(win));
+            var canSpell = InlineSpellCheckerUI.canSpellCheck;
+            InlineSpellCheckerUI.initFromEvent(rangeParent, rangeOffset);
+            this.showItem("spell-check-enabled", canSpell);
+            this.showItem("spell-separator", canSpell);
+            return;
+          }
+        }
+
         // See if the user clicked on an image.
         if ( this.target.nodeType == Node.ELEMENT_NODE ) {
             if ( this.target instanceof Components.interfaces.nsIImageLoadingContent && this.target.currentURI  ) {
@@ -294,10 +362,20 @@ nsContextMenu.prototype = {
                 if ( this.target.ownerDocument instanceof ImageDocument )
                    this.onStandaloneImage = true;
             } else if ( this.target instanceof HTMLInputElement ) {
-               type = this.target.getAttribute("type");
-               this.onTextInput = this.isTargetATextBox(this.target);
+                this.onTextInput = this.isTargetATextBox(this.target);
+                // allow spellchecking UI on all writable text boxes except passwords
+                if (!this.target.readOnly && !this.target.disabled && this.target.type == "text") {
+                    this.possibleSpellChecking = true;
+                    InlineSpellCheckerUI.init(this.target.QueryInterface(Components.interfaces.nsIDOMNSEditableElement).editor);
+                    InlineSpellCheckerUI.initFromEvent(rangeParent, rangeOffset);
+                }
             } else if ( this.target instanceof HTMLTextAreaElement ) {
-                 this.onTextInput = true;
+                this.onTextInput = true;
+                if (!this.target.readOnly && !this.target.disabled) {
+                    this.possibleSpellChecking = true;
+                    InlineSpellCheckerUI.init(this.target.QueryInterface(Components.interfaces.nsIDOMNSEditableElement).editor);
+                    InlineSpellCheckerUI.initFromEvent(rangeParent, rangeOffset);
+                }
             } else if ( this.target instanceof HTMLHtmlElement ) {
                // pages with multiple <body>s are lame. we'll teach them a lesson.
                var bodyElt = this.target.ownerDocument.getElementsByTagName("body")[0];
@@ -521,14 +599,14 @@ nsContextMenu.prototype = {
         openNewWindowWith( this.linkURL(), true );
     },
     // Open linked-to URL in a new tab.
-    openLinkInTab : function () {
+    openLinkInTab : function ( reverseBackgroundPref ) {
         // Determine linked-to URL.
-        openNewTabWith( this.linkURL(), true, false );
+        openNewTabWith( this.linkURL(), true, reverseBackgroundPref );
     },
     // Open frame in a new tab.
-    openFrameInTab : function () {
+    openFrameInTab : function ( reverseBackgroundPref ) {
         // Determine linked-to URL.
-        openNewTabWith( this.target.ownerDocument.location.href );
+        openNewTabWith( this.target.ownerDocument.location.href, true, reverseBackgroundPref );
     },
     // Reload clicked-in frame.
     reloadFrame : function () {
@@ -627,35 +705,30 @@ nsContextMenu.prototype = {
         saveImageURL( this.imageURL, null, "SaveImageTitle", false,
                       getReferrer(document) );
     },
-    // Generate email address and put it on clipboard.
-    copyEmail : function () {
-        // Copy the comma-separated list of email addresses only.
+    // Generate email address.
+    getEmail : function () {
+        // Get the comma-separated list of email addresses only.
         // There are other ways of embedding email addresses in a mailto:
         // link, but such complex parsing is beyond us.
-        var url = this.linkURL();
-        var qmark = url.indexOf( "?" );
         var addresses;
-        
-        if ( qmark > 7 ) {                   // 7 == length of "mailto:"
-            addresses = url.substring( 7, qmark );
-        } else {
-            addresses = url.substr( 7 );
-        }
-
-        // Let's try to unescape it using a character set
         try {
+          // Let's try to unescape it using a character set
           var characterSet = this.target.ownerDocument.characterSet;
           const textToSubURI = Components.classes["@mozilla.org/intl/texttosuburi;1"]
                                          .getService(Components.interfaces.nsITextToSubURI);
+          addresses = this.linkURL().match(/^mailto:([^?]+)/)[1];
           addresses = textToSubURI.unEscapeURIForUI(characterSet, addresses);
         }
         catch(ex) {
           // Do nothing.
         }
-
+        return addresses;
+    },
+    // Copy email to clipboard
+    copyEmail : function () {
         var clipboard = this.getService( "@mozilla.org/widget/clipboardhelper;1",
                                          Components.interfaces.nsIClipboardHelper );
-        clipboard.copyString(addresses);
+        clipboard.copyString(this.getEmail());
     },    
     addBookmark : function() {
       var docshell = document.getElementById( "content" ).webNavigation;
@@ -819,6 +892,11 @@ nsContextMenu.prototype = {
         searchStr = searchStr.replace(/\s+/g, " ");
         return searchStr;
     },
+
+    // Returns true if anything is selected.
+    isContentSelection: function() {
+        return !document.commandDispatcher.focusedWindow.getSelection().isCollapsed;
+    },
     
     // Convert relative URL to absolute, using document's <base>.
     makeURLAbsolute : function ( base, url ) {
@@ -860,6 +938,16 @@ nsContextMenu.prototype = {
         }
       }
       return false;  
+    },
+
+    addDictionaries : function()
+    {
+      try {
+        var url = pref.getComplexValue("editor.spellcheckers.url",
+                                       Components.interfaces.nsIPrefLocalizedString).data;
+        window.openDialog(getBrowserURL(), "_blank", "chrome,all,dialog=no", url);
+      }
+      catch (ex) {}
     }
 };
 
